#include "FLoPconfig.h"
#include "comm.h"
#include "ctrl.h"
#include "flopdebug.h"
#include <signal.h>
#include "Signal.h"
#include "status.h"
#include <pthread.h>
#include <sys/select.h>
#include <time.h>

/* global values defined elsewhere */

#ifdef DROP
extern u_int LowWater;
extern u_int HighWater;
#endif

#ifdef DEBUGFLOP
extern u_int8_t DebugVector[DEBUG_MAX];
#endif

extern u_int AlarmDelay;

#ifdef SERVSOCK
#include "socket_database.h"
void DumpConfig();
extern DatabaseData *DBdata;
extern u_int UTC;              /* Timezone, 1 means use UTC */
#ifdef ALERT
extern int UnixSocketPriority;
extern int ReverseUnixPriority;
#endif
#endif
void Statistics();


/* 
 * Function: ControlThread
 *
 * Purpose: reads commands from a FIFO to change runtime variables
 *  
 * Arguments: ptr to the fd of the opened FIFO
 *
 * Returns: no return 
 *
 * Author: DG
 */

void *
ControlThread(void *ptr) 
{
  int s;
  u_int TimeDelay, interval=1;
  time_t LastTime;
  time_t ActualTime;
  int result;
  static char buffer[STRLEN];
  ssize_t n;
  char *key, *val;
  fd_set rfds;
  struct timeval tv;

  s=*(int *) ptr;

  FD_ZERO(&rfds);

  LastTime=time(NULL);

  while (1)
   {
     key=NULL;
     val=NULL;

     FD_SET(s,&rfds);
    
     if (AlarmDelay)
       {
	 /* use select with a timeout */
	 /* reset time interval to original one, this may have been
	  * changed if either select times out or data arrives
	  */
	 tv.tv_sec = interval;
	 tv.tv_usec = 0;

	 result = select(s+1,&rfds, NULL, NULL, &tv);
       }
     else
       {
	 /* no periodic job has to be done, wait until data arrives */
	 result = select(s+1,&rfds, NULL, NULL, NULL);
       }

     if (result ==-1)
       {
	 DG_strerror_r(errno,buffer,STRLEN);
	 Printf("CTRL> Error with select: %s, ignoring\n",buffer);
	 continue;
       }

     if (result ==0)
       {
	 ActualTime=time(NULL);
	 TimeDelay=ActualTime-LastTime;
	 if (AlarmDelay && AlarmDelay <= TimeDelay)
	   {
	     TimeDelay=0;
	     LastTime=ActualTime;
	     /* get the statistics */
	     Statistics();
	   }
	 /* go back to select */
	 continue;
       }
     
     if ((n=read(s,buffer,STRLEN)) >0 )
       {
	 if (buffer[n-1]=='\n') n--;
	 buffer[n]='\0';
	 DEBUGF(DebugOutput(DEBUG_CTRL,5,"Input: %s\n",buffer););

         if (strncasecmp(buffer,"EXIT thread",11)==0)
           {
	      DEBUGF(DebugOutput(DEBUG_CTRL,5,"CTRL> Got exit: %s\n",buffer););
	      close(s);
	      pthread_exit(NULL);
            }
       }    
     else
       {
         if (n==-1)
            {
              DG_strerror_r(errno,buffer,STRLEN);
              DEBUGF(DebugOutput(DEBUG_CTRL,5,"CTRL> Read error: %s\n",buffer););
              if (errno == EINTR || errno == EAGAIN) continue;
            }
         DEBUGF(DebugOutput(DEBUG_CTRL,5,"invalid read of bytes: %i\n",n););
         continue;
       }
     if ( (strstr(buffer,":") == NULL) && (strstr(buffer,"=") == NULL) )
        {
           Printf("CT> %s (missing : or = ?)\n",buffer);
        }
     else
       {
	 key = strtok(buffer,":=");
	 /* second argument is the value, remove tabs */
	 val = strtok(NULL,"\0");
	 /* remove leading spaces or tabs */
	 while ( key[0]==' ' || key[0]=='\t') key++;
	 while ( val[0]==' ' || val[0]=='\t') val++;
	 /* remove spaces at end of value */
	 while ( (strlen(key) > 1) && (key[strlen(key)-1] == ' ') )
                    key[strlen(key)-1]='\0';
	 while ( (strlen(val) > 1) && (val[strlen(val)-1] == ' ') )
	   val[strlen(val)-1]='\0';
	 DEBUGF(DebugOutput(DEBUG_CTRL,5,"Key: <%s> val: <%s>\n",key,val););
	 SetValues(key,val);
       }
   }
  DEBUGF(DebugOutput(DEBUG_CTRL,5,"ControlThread> Done."););
  close(s);
  pthread_exit(NULL);

}

void 
SetValues(char *arg, char *val)
{
  u_int u;
  char *c;
  int i;
  u_int DebugArea,DebugLevel;

  DEBUGF(DebugOutput(DEBUG_SV,5,"SetValues> Input: <%s> : <%s>\n",arg,val););

  if ( !strncasecmp(arg,"statistics",10) )
    {
      if (atoi(val))
	{ 
          Statistics();
	}
      return;
    }
  if ( !strncasecmp(arg,"alarmdelay",10) )
    {
      i=0;
      if (AlarmDelay==0)
	{
	  i=1;
	}

      i=atoi(val);
      if ( i > 0) AlarmDelay=i;
      else AlarmDelay=0;

      Printf("AlarmDelay is now %i\n",AlarmDelay);
      return;
    }

  if ( !strncasecmp(arg,"LowWater",8) )
    {

#ifdef DROP
      u=atoi(val);
      if (u > HighWater) 
	{
	  Printf("Please increase HighWater(%u) first\n",HighWater);
	}
      else
	{
	  LowWater=u; 
	}
      Printf("LowWater now : %u\n",LowWater);

#else
      Printf("Option LowWater not enabled!\n");
      Printf("Recompile with -DDROP\n");
#endif
      return;
    }

  if ( !strncasecmp(arg,"HighWater",9) )
    {

#ifdef DROP
      u=atoi(val);
      if (u < LowWater) 
	{
	  Printf("Please decrease LowWater(%u) first\n",LowWater);
	}
      else
	{
	  HighWater=u; 
	}
      Printf("HighWater now : %u\n",HighWater);
       
#else
      Printf("Option HighWater not enabled!\n");
      Printf("Recompile with -DDROP\n");
#endif
      return;
    }

  if ( !strncasecmp(arg,"DEBUG",5) )
    {
#ifdef DEBUGFLOP
      c=strchr(val,',');
      if (c == NULL)
	{
	  DebugArea=atoi(val);
	  DebugLevel=9;
	}
      else
	{
	  *c='\0';c++;
	  DebugArea=atoi(val);
	  DebugLevel=atoi(c);
	}

	  /* DebugLevel has maximum of 9 */
      if (DebugLevel > 9 ) DebugLevel=9;

          /* Set debug value for area */
      if ( (DebugArea < DEBUG_MAX) && (DebugArea > 0) )
	{	    
	  if (DebugVector[DebugArea]==0)
	    {
	      DebugVector[DebugArea]=DebugLevel;
	      DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
				 "DebugArea: %u, DebugLevel: %u\n",
				 DebugArea,DebugLevel););
	    }
	}
	  
      /* if area is ascii "ALL" then set all values ( if not 
       *  already set to an value) to DebugLevel 
       */
      if ( DebugArea == 0 && (!strncasecmp(val,"ALL",3)) )
	{
	  DEBUGF(DebugOutput(DEBUG_CMDLINE,1,
			     "DebugArea: ALL, DebugLevel: %u\n",
			     DebugLevel););
	  for (i=0;i<DEBUG_MAX;i++)
	    {
	      if (DebugVector[i] == 0 ) DebugVector[i]=DebugLevel;
	    }
	}
#else
      Printf("Debugging is disabled. Please reconfigure FLoP with "
	     "debugging enabled.\n");
#endif
      return;
    }
#ifdef SERVSOCK
      if ( !strncasecmp(arg,"dbtrust",7) )
        {
          DBdata->DBtrust=atoi(val);
	  if (DBdata->DBtrust)
	    Printf("DBtrust enabled\n");
	  else
	    Printf("DBtrust disabled\n");
          return;
        }

      if ( !strncasecmp(arg,"TimeZone",8) )
	{
          UTC=atoi(val);
	  if(UTC)
	    Printf("Use of UTC time is enabled\n");
	  else
	    Printf("Use of UTC time is disabled\n");
	  return;
	}

      if ( !strncasecmp(arg,"DumpConfig",8) )
	{
          if(atoi(val))
	    { 
	      Printf("Dumping Config for pid: %i\n",(int ) getpid());
	      DumpConfig();
	    }
	  return;
	}

      if ( (!strncasecmp(arg,"UnixSocketPriority",18)) || 
	   (!strncasecmp(arg,"Priority",8)))
	{

#ifdef ALERT
	  UnixSocketPriority=atoi(val);
	  if (UnixSocketPriority < 0)
	    {
	      Printf("Reverting unix socket priority for alert\n");
	      ReverseUnixPriority = 1;
	      UnixSocketPriority = abs(UnixSocketPriority);
	    }
	  else
	    {
	      Printf("Normal (old) unix socket priority for alert\n");
	      ReverseUnixPriority = 0;
	    }
	  Printf("UnixSocketPriority for alerts is set to: %i\n",
		 UnixSocketPriority);
#else
          Printf("Option UnixSocketPriority not enabled!\n");
	  Printf("Recompile with -DALERT\n");
#endif
	  return;
	}

#endif
  Printf("\nWARNING: ignoring unknown option <%s>: <%s>\n",arg,val);

}
